import datetime
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from vesta import __version__ as version

# Project

project = "Vesta"
copyright = f"{datetime.date.today().year} Jon Parise"
author = "Jon Parise"
release = version

# General

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_preserve_defaults = True
autodoc_type_aliases = {
    "Row": "vesta.chars.Row",
    "Rows": "vesta.chars.Rows",
}


# HTML

html_theme = "furo"
html_title = f"Vesta Documentation ({version})"
html_logo = "_static/logo.png"
html_static_path = ["_static"]
html_copy_source = False
html_show_sourcelink = False
html_show_sphinx = False



import datetime
from typing import Callable
from typing import Collection
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from .clients import Client

Result = Union[str, List[List[int]]]
Runnable = Callable[[], Result]
# Callback = Callable[[], CallbackResult]
# Runnable = Callable[[], Callback]

Day = Literal[0, 1, 2, 3, 4, 5, 6]


class Job:
    def __init__(
        self,
        func: Runnable,
        when: datetime.time,
        period: Optional[datetime.timedelta] = None,
        until: Optional[datetime.time] = None,
        days: Collection[Day] = (0, 1, 2, 3, 4, 5, 6),
    ):
        self.func = func
        self.when = when
        self.period = period
        self.until = until
        self.days = days
        self.next_run = self._schedule()

    def __lt__(self, other) -> bool:
        return self.next_run < other.next_run

    def _schedule(self) -> datetime.datetime:
        next_run = datetime.datetime.now()

        if next_run.time() < self.when:
            next_run = datetime.datetime.combine(next_run.date(), self.when)
        elif self.period is not None:
            next_run += self.period

        # If we've run out of scheduled time for today, advance to tomorrow.
        if (self.period is None and next_run.time() > self.when) or (
            self.until is not None and next_run.time() > self.until
        ):
            tomorrow = next_run.date() + datetime.timedelta(days=1)
            next_run = datetime.datetime.combine(tomorrow, self.when)

        # We may also need to advance to the next available day.
        while next_run.weekday not in self.days:
            next_day = next_run.date() + datetime.timedelta(days=1)
            next_run = datetime.datetime.combine(next_day, self.when)

        return next_run

    def run(self) -> Result:
        result = self.func()
        self.next_run = self._schedule()
        return result


class Schedule:
    def __init__(self):
        self.jobs: List[Job] = []

    def add(
        self,
        func: Runnable,
        when: Optional[datetime.time] = None,
        *,
        period: Optional[datetime.timedelta] = None,
        until: Optional[datetime.time] = None,
        days: Collection[Day] = (0, 1, 2, 3, 4, 5, 6),
    ):
        if when is None:
            when = datetime.time()
        if until is not None and until <= when:
            raise ValueError("'until' must be after 'when'")
        if len(days) < 1:
            raise ValueError("must specify at least one day")

        # if bool(at) != bool(every):
        #     raise ValueError("must specify one of 'at' or 'every'")
        # self.jobs.append(Job(func))

    def run(self, client: Client):
        ...



import pytest
from requests_mock import Mocker

from vesta.chars import COLS
from vesta.chars import ROWS
from vesta.clients import Client
from vesta.clients import LocalClient
from vesta.clients import ReadWriteClient


@pytest.fixture
def client():
    return Client("key", "secret")


@pytest.fixture
def local_client():
    return LocalClient("key")


@pytest.fixture
def rw_client():
    return ReadWriteClient("key")


class TestClient:
    def test_base_url(self):
        base_url = "https://www.example.com"
        client = Client("key", "secret", base_url=base_url)
        assert client.session.base_url == base_url
        assert base_url in repr(client)

    def test_headers(self):
        client = Client("key", "secret", headers={"User-Agent": "Vesta"})
        assert client.session.headers["X-Vestaboard-Api-Key"] == "key"
        assert client.session.headers["X-Vestaboard-Api-Secret"] == "secret"
        assert client.session.headers["User-Agent"] == "Vesta"

    def test_get_subscriptions(self, client: Client, requests_mock: Mocker):
        subscriptions = [True]
        requests_mock.get(
            "https://platform.vestaboard.com/subscriptions",
            json={"subscriptions": subscriptions},
        )
        assert client.get_subscriptions() == subscriptions

    def test_get_viewer(self, client: Client, requests_mock: Mocker):
        viewer = {"_id": 1}
        requests_mock.get("https://platform.vestaboard.com/viewer", json=viewer)
        assert client.get_viewer() == viewer

    def test_post_message_text(self, client: Client, requests_mock: Mocker):
        text = "abc"

        def match_json(request) -> bool:
            return request.json() == {"text": text}

        requests_mock.post(
            "https://platform.vestaboard.com/subscriptions/sub_id/message",
            additional_matcher=match_json,
            json={},
        )
        client.post_message("sub_id", text)

    def test_post_message_list(self, client: Client, requests_mock: Mocker):
        chars = [[0] * COLS] * ROWS

        def match_json(request) -> bool:
            return request.json() == {"characters": chars}

        requests_mock.post(
            "https://platform.vestaboard.com/subscriptions/sub_id/message",
            additional_matcher=match_json,
            json={},
        )
        client.post_message("sub_id", chars)

    def test_post_message_list_dimensions(self, client: Client):
        with pytest.raises(ValueError, match=rf"expected a \({ROWS}, {COLS}\) array"):
            client.post_message("sub_id", [])

    def test_post_message_type(self, client: Client):
        with pytest.raises(TypeError, match=r"unsupported message type"):
            client.post_message("sub_id", True)  # type: ignore


class TestLocalClient:
    def test_base_url(self):
        base_url = "http://example.local"
        local_client = LocalClient("key", base_url=base_url)
        assert local_client.session.base_url == base_url
        assert base_url in repr(local_client)

    def test_api_key_init(self):
        local_api_key = "key"
        local_client = LocalClient(local_api_key)
        assert local_client.enabled
        assert local_client.api_key == local_api_key

    def test_api_key_property(self):
        local_client = LocalClient()
        assert not local_client.enabled
        assert local_client.api_key is None

        local_api_key = "key"
        local_client.api_key = local_api_key
        assert local_client.enabled
        assert local_client.api_key == local_api_key

    def test_enable(self, requests_mock: Mocker):
        local_client = LocalClient()
        local_api_key = "key"
        requests_mock.post(
            "http://vestaboard.local:7000/local-api/enablement",
            json={"apiKey": local_api_key},
        )
        rv = local_client.enable("enablement_token")
        assert rv == local_api_key
        assert local_client.api_key == local_api_key
        assert local_client.enabled
        assert requests_mock.last_request
        assert (
            requests_mock.last_request.headers.get(
                "X-Vestaboard-Local-Api-Enablement-Token"
            )
            == "enablement_token"
        )

    def test_enable_failure(self, requests_mock: Mocker):
        local_client = LocalClient()
        requests_mock.post("http://vestaboard.local:7000/local-api/enablement")
        rv = local_client.enable("enablement_token")
        assert rv is None
        assert not local_client.api_key
        assert not local_client.enabled

    def test_not_enabled(self):
        local_client = LocalClient()
        with pytest.raises(RuntimeError, match=r"Local API has not been enabled"):
            local_client.read_message()
        with pytest.raises(RuntimeError, match=r"Local API has not been enabled"):
            local_client.write_message([])

    def test_read_message(self, local_client: LocalClient, requests_mock: Mocker):
        chars = [[0] * COLS] * ROWS
        requests_mock.get(
            "http://vestaboard.local:7000/local-api/message", json={"message": chars}
        )
        message = local_client.read_message()
        assert message == chars

    def test_read_message_empty(self, local_client: LocalClient, requests_mock: Mocker):
        requests_mock.get("http://vestaboard.local:7000/local-api/message")
        message = local_client.read_message()
        assert message is None

    def test_write_message(self, local_client: LocalClient, requests_mock: Mocker):
        chars = [[0] * COLS] * ROWS
        requests_mock.post(
            "http://vestaboard.local:7000/local-api/message", status_code=201
        )
        assert local_client.write_message(chars)
        assert requests_mock.last_request
        assert requests_mock.last_request.json() == chars

    def test_write_message_dimensions(self, local_client: LocalClient):
        with pytest.raises(ValueError, match=rf"expected a \({ROWS}, {COLS}\) array"):
            local_client.write_message([])


class TestReadWriteClient:
    def test_base_url(self):
        base_url = "http://example.local"
        rw_client = ReadWriteClient("key", base_url=base_url)
        assert rw_client.session.base_url == base_url
        assert base_url in repr(rw_client)

    def test_headers(self):
        client = ReadWriteClient("key", headers={"User-Agent": "Vesta"})
        assert client.session.headers["X-Vestaboard-Read-Write-Key"] == "key"
        assert client.session.headers["User-Agent"] == "Vesta"

    def test_read_message(self, rw_client: ReadWriteClient, requests_mock: Mocker):
        chars = [[0] * COLS] * ROWS
        requests_mock.get(
            "https://rw.vestaboard.com",
            json={"currentMessage": {"layout": chars}},
        )
        message = rw_client.read_message()
        assert message == chars

    def test_read_message_empty(
        self, rw_client: ReadWriteClient, requests_mock: Mocker
    ):
        requests_mock.get("https://rw.vestaboard.com")
        message = rw_client.read_message()
        assert message is None

    def test_write_message(self, rw_client: ReadWriteClient, requests_mock: Mocker):
        chars = [[0] * COLS] * ROWS
        requests_mock.post("https://rw.vestaboard.com", status_code=201)
        assert rw_client.write_message(chars)
        assert requests_mock.last_request
        assert requests_mock.last_request.json() == chars

    def test_write_message_dimensions(self, rw_client: ReadWriteClient):
        with pytest.raises(ValueError, match=rf"expected a \({ROWS}, {COLS}\) array"):
            rw_client.write_message([])



from .chars import Color
from .chars import encode
from .chars import encode_row
from .chars import encode_text
from .chars import pprint
from .clients import Client
from .clients import LocalClient
from .clients import ReadWriteClient

__all__ = (
    "Color",
    "encode",
    "encode_row",
    "encode_text",
    "pprint",
    "Client",
    "LocalClient",
    "ReadWriteClient",
)

__version__ = "0.8.0"



# Copyright (c) 2021-2022 Jon Parise <jon@indelible.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import enum
import math
import sys
from typing import Container
from typing import List
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import Union
from typing import cast

ROWS = 6
COLS = 22
PRINTABLE = " ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890!@#$() - +&=;: '\"%,.  /? °"
CHARMAP = {c: i for i, c in enumerate(PRINTABLE) if i == 0 or c != " "}

#: A row of character codes.
Row = List[int]

#: A list of rows, forming a character grid.
Rows = List[Row]


def validate_rows(rows: Rows) -> None:
    """Check if ``rows`` is a well-formed grid of character codes.

    :raises ValueError: if ``rows`` does not have the correct dimensions
    """
    if len(rows) != ROWS or not all(len(row) == COLS for row in rows):
        raise ValueError(f"expected a ({ROWS}, {COLS}) array of encoded characters")


class Color(enum.IntEnum):
    """Color chips"""

    ansi: str

    def __new__(cls, value: int, ansi: str) -> "Color":
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.ansi = ansi
        return obj

    # fmt: off
    BLACK   = (0,  "\033[0m")   # noqa: E221
    RED     = (63, "\033[31m")  # noqa: E221
    ORANGE  = (64, "\033[33m")  # noqa: E221
    YELLOW  = (65, "\033[93m")  # noqa: E221
    GREEN   = (66, "\033[32m")  # noqa: E221
    BLUE    = (67, "\033[94m")  # noqa: E221
    VIOLET  = (68, "\033[95m")  # noqa: E221
    WHITE   = (69, "\033[97m")  # noqa: E221
    # fmt: on


# The set of all supported character codes.
CHARCODES = frozenset(CHARMAP.values()).union(Color)


def encode(s: str) -> Row:
    """Encodes a string as a list of character codes.

    In addition to printable characters, the string can contain character code
    sequences inside curly braces, such as ``{5}`` or ``{65}``.

    :raises ValueError: if the string contains unsupported characters or codes

    >>> encode("{67} Hello, World {68}")
    [67, 0, 8, 5, 12, 12, 15, 55, 0, 23, 15, 18, 12, 4, 0, 68]
    """
    out = []
    lens = len(s)
    skip_to = 0
    for i, c in enumerate(s.upper()):
        if i < skip_to:
            continue

        if c == "{":
            if lens > i + 2 and s[i + 2] == "}":
                out.append(int(s[i + 1]))
                skip_to = i + 3
            elif lens > i + 3 and s[i + 3] == "}":
                out.append(int(s[i + 1 : i + 3]))
                skip_to = i + 4
            else:
                raise ValueError(f"{i+1}: missing }} at index {i+2} or {i+3}")
            if out[-1] not in CHARCODES:
                raise ValueError(f"{i+2}: unsupported character code: {out[-1]}")
        else:
            try:
                out.append(CHARMAP[c])
            except KeyError:
                raise ValueError(f"{i+1}: unsupported character: {c}")

    return out


def encode_row(s: str, align: str = "left", fill: int = Color.BLACK) -> Row:
    """Encodes a string as a row of character codes.

    In addition to printable characters, the string can contain character code
    sequences inside curly braces, such as ``{5}`` or ``{65}``.

    ``align`` controls the text's alignment within the row: `left`, `right`, or
    `center`. The ``fill`` character code (generally a :py:class:`Color`) is
    used to fill out any additional space.

    :raises ValueError: if the string contains unsupported characters or codes,
                        or if the resulting encoding sequence would exceed the
                        maximum number of supported columns

    >>> encode_row("{67} Hello, World {68}", align="center")
    [0, 0, 0, 67, 0, 8, 5, 12, 12, 15, 55, 0, 23, 15, 18, 12, 4, 0, 68, 0, 0, 0]
    """
    row = encode(s)

    if len(row) > COLS:
        raise ValueError(f"{s!r} results in {len(row)} characters (max {COLS})")

    return _format_row(row, align, 0, int(fill))


def encode_text(
    s: str,
    align: str = "left",
    valign: Optional[str] = "top",
    margin: int = 0,
    fill: int = Color.BLACK,
    breaks: Container[int] = frozenset({0}),
) -> Rows:
    """Encodes a string of text into rows of character codes.

    In addition to printable characters, the string can contain character code
    sequences inside curly braces, such as ``{5}`` or ``{65}``.

    ``align`` controls the text's alignment within the row: `left`, `right`, or
    `center`. ``valign`` controls the text's vertical alignment within the full
    board: `top`, `middle`, `bottom`, or None (to never add rows, potentially
    resulting in a partial board).

    ``margin`` specifies the width (in columns) of the left and right margins.
    The ``fill`` character code (generally a :py:class:`Color`) is used to fill
    out any additional space.

    ``breaks`` is the set of character codes used to compute line breaks. If a
    line of text won't fit in the available columns, it will be "broken" at the
    first preceding break character, and the remaining characters will continue
    on the next row (potentially subject to additional breaks). If a break
    cannot be found, the line will be broken at the column limit (potentially
    mid-"word").

    :raises ValueError: if the string contains unsupported characters or codes,
                        or if the resulting encoding sequence would exceed the
                        maximum number of supported rows

    >>> encode_text('multiple\\nlines\\nof\\ntext', align="center", valign="middle")
    ... # doctest: +NORMALIZE_WHITESPACE
    [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 13, 21, 12, 20, 9, 16, 12, 5, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 12, 9, 14, 5, 19, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 20, 5, 24, 20, 0, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
    """
    fill = int(fill)
    max_cols = COLS - margin * 2
    rows: Rows = []

    def find_break(line: Row) -> Tuple[int, int]:
        end = min(len(line), max_cols)
        for pos in range(end, 0, -1):
            if line[pos] in breaks:
                return pos, pos + 1
        return end, end

    for line in map(encode, s.splitlines()):
        while len(line) > max_cols:
            pos, resume = find_break(line)
            rows.append(_format_row(line[:pos], align, margin, fill))
            line = line[resume:]

        rows.append(_format_row(line, align, margin, fill))

    nrows = len(rows)
    if nrows < ROWS and valign is not None:
        empty = [fill] * COLS
        if valign == "top":
            rows += [empty] * (ROWS - nrows)
        elif valign == "bottom":
            rows = [empty] * (ROWS - nrows) + rows
        elif valign == "middle":
            pad = (ROWS - nrows) / 2
            rows = [empty] * math.floor(pad) + rows + [empty] * math.ceil(pad)
        else:
            raise ValueError(f"unknown vertical alignment: {valign}")
    elif nrows > ROWS:
        raise ValueError(f"{s!r} results in {nrows} lines (max {ROWS})")

    return rows


def _format_row(row: Row, align: str, margin: int, fill: int) -> Row:
    assert len(row) <= COLS - margin * 2

    if align == "left":
        row = [fill] * margin + row + [fill] * (COLS - len(row) - margin)
    elif align == "right":
        row = [fill] * (COLS - len(row) - margin) + row + [fill] * margin
    elif align == "center":
        pad = (COLS - len(row)) / 2
        row = [fill] * math.floor(pad) + row + [fill] * math.ceil(pad)
    else:
        raise ValueError(f"unknown alignment: {align}")

    assert len(row) == COLS

    return row


def pprint(
    data: Union[Row, Rows],
    stream: TextIO = sys.stdout,
    *,
    sep: str = "|",
    block: str = "◼︎",
) -> None:
    """Prints a console-formatted representation of encoded character data.

    ``data`` may be a single list or a two-dimensional array of character codes.
    """
    rows = cast(Rows, data if data and isinstance(data[0], list) else [data])

    # Assume all TTYs support color.
    colors = stream.isatty()
    if colors and sep:
        sep = f"\033[90m{sep}\033[0m"

    def symbol(code: int) -> str:
        if 0 <= code < len(PRINTABLE):
            return PRINTABLE[code]
        try:
            color = Color(code)  # type: ignore
        except ValueError:
            raise ValueError(f"unknown character code: {code}")
        return f"{color.ansi}{block}\033[0m" if colors else block

    lines = [sep + sep.join(map(symbol, row)) + sep for row in rows]
    stream.write("\n".join(lines) + "\n")



# Copyright (c) 2021-2022 Jon Parise <jon@indelible.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union
from typing import cast
from urllib.parse import urljoin

import requests

from .chars import Rows
from .chars import validate_rows

__all__ = ["Client", "LocalClient"]


class Session(requests.Session):
    def __init__(self, base_url: str):
        super().__init__()
        self.base_url = base_url

    def request(
        self,
        method: Union[str, bytes],
        url: Union[str, bytes],
        *args,
        **kwargs,
    ) -> requests.Response:
        url = urljoin(self.base_url, url if isinstance(url, str) else url.decode())
        return super().request(method, url, *args, **kwargs)


class Client:
    """Provides a Vestaboard API client interface.

    Credentials must be provided as an ``api_key`` and ``api_secret``.

    Optionally, an alternate ``base_url`` can be specified, as well as any
    additional HTTP ``headers`` that should be sent with every request
    (such as a custom `User-Agent` header).
    """

    def __init__(
        self,
        api_key: str,
        api_secret: str,
        *,
        base_url: str = "https://platform.vestaboard.com",
        headers: Optional[Mapping[str, str]] = None,
    ):
        self.session = Session(base_url)
        self.session.headers["X-Vestaboard-Api-Key"] = api_key
        self.session.headers["X-Vestaboard-Api-Secret"] = api_secret
        if headers:
            self.session.headers.update(headers)

    def __repr__(self):
        return f"{type(self).__name__}(base_url={self.session.base_url!r})"

    def get_subscriptions(self) -> List[Dict[str, Any]]:
        """Lists all subscriptions to which the viewer has access."""
        r = self.session.get("/subscriptions")
        r.raise_for_status()
        return r.json().get("subscriptions", [])

    def get_viewer(self) -> Dict[str, Any]:
        """Describes the currently authenticated viewer."""
        r = self.session.get("/viewer")
        r.raise_for_status()
        return r.json()

    def post_message(
        self,
        subscription_id: str,
        message: Union[str, Rows],
    ) -> Dict[str, Any]:
        """Post a new message to a subscription.

        The authenticated viewer must have access to the subscription.

        `message` can be either a string of text or a two-dimensional (6, 22)
        array of character codes representing the exact positions of characters
        on the board.

        If text is specified, the lines will be centered horizontally and
        vertically. Character codes will be inferred for alphanumeric and
        punctuation characters, or they can be explicitly specified using curly
        braces containing the character code (such as ``{5}`` or ``{65}``).

        :raises ValueError: if ``message`` is a list with unsupported dimensions
        """
        data: Dict[str, Union[str, Rows]]
        if isinstance(message, str):
            data = {"text": message}
        elif isinstance(message, list):
            validate_rows(message)
            data = {"characters": message}
        else:
            raise TypeError(f"unsupported message type: {type(message)}")

        r = self.session.post(
            f"/subscriptions/{subscription_id}/message",
            json=data,
        )
        r.raise_for_status()
        return r.json()


class LocalClient:
    """Provides a Vestaboard Local API client interface.

    A Local API key is required to read or write messages. This key is obtained
    by enabling the Vestaboard's Local API using a Local API Enablement Token.

    If you've already enabled your Vestaboard's Local API, that key can be
    provided immediately. Otherwise, it can be set after the client is
    constructed by calling :py:meth:`~enable`, which also returns the Local API
    key for future reuse.

    An alternate ``base_url`` can also be specified.

    .. versionadded:: 0.8.0
    """

    def __init__(
        self,
        local_api_key: Optional[str] = None,
        *,
        base_url: str = "http://vestaboard.local:7000",
    ):
        self.session = Session(base_url)
        if local_api_key is not None:
            self.api_key = local_api_key

    def __repr__(self):
        return f"{type(self).__name__}(base_url={self.session.base_url!r})"

    @property
    def api_key(self) -> Optional[str]:
        """The client's Local API key."""
        return cast(
            Optional[str],
            self.session.headers.get("X-Vestaboard-Local-Api-Key"),
        )

    @api_key.setter
    def api_key(self, value: str) -> None:
        self.session.headers["X-Vestaboard-Local-Api-Key"] = value

    @property
    def enabled(self) -> bool:
        """Check if :py:attr:`~api_key` has been set, indicating that Local API
        support has been enabled."""
        return self.api_key is not None

    def enable(self, enablement_token) -> Optional[str]:
        """Enable the Vestaboard's Local API using a Local API Enablement Token.

        If successful, the Vestaboard's Local API key will be returned and the
        client's :py:attr:`~api_key` property will be updated to the new value.
        """
        r = self.session.post(
            "/local-api/enablement",
            headers={"X-Vestaboard-Local-Api-Enablement-Token": enablement_token},
        )
        r.raise_for_status()

        try:
            local_api_key = r.json().get("apiKey")
        except requests.JSONDecodeError:
            local_api_key = None

        if local_api_key:
            self.api_key = local_api_key

        return local_api_key

    def read_message(self) -> Optional[Rows]:
        """Read the Vestaboard's current message."""
        if not self.enabled:
            raise RuntimeError("Local API has not been enabled")
        r = self.session.get("/local-api/message")
        r.raise_for_status()
        try:
            return r.json().get("message")
        except requests.JSONDecodeError:
            return None

    def write_message(self, message: Rows) -> bool:
        """Write a message to the Vestaboard.

        `message` must be a two-dimensional (6, 22) array of character codes
        representing the exact positions of characters on the board.

        :raises ValueError: if ``message`` is a list with unsupported dimensions
        """
        if not self.enabled:
            raise RuntimeError("Local API has not been enabled")
        validate_rows(message)
        r = self.session.post("/local-api/message", json=message)
        r.raise_for_status()
        return r.status_code == requests.codes.CREATED


class ReadWriteClient:
    """Provides a Vestaboard Read / Write API client interface.

    A Read / Write API key is required to read or write messages. This key is
    obtained by enabling the Vestaboard's Read / Write API via the Settings
    section of the mobile app or from the Developer section of the web app.

    Optionally, an alternate ``base_url`` can be specified, as well as any
    additional HTTP ``headers`` that should be sent with every request
    (such as a custom `User-Agent` header).

    .. versionadded:: 0.8.0
    """

    def __init__(
        self,
        read_write_key: str,
        base_url: str = "https://rw.vestaboard.com/",
        headers: Optional[Mapping[str, str]] = None,
    ):
        self.session = Session(base_url)
        self.session.headers["X-Vestaboard-Read-Write-Key"] = read_write_key
        if headers:
            self.session.headers.update(headers)

    def __repr__(self):
        return f"{type(self).__name__}(base_url={self.session.base_url!r})"

    def read_message(self) -> Optional[Rows]:
        """Read the Vestaboard's current message."""
        r = self.session.get("")
        r.raise_for_status()
        try:
            return r.json().get("currentMessage", {}).get("layout")
        except requests.JSONDecodeError:
            return None

    def write_message(self, message: Rows) -> bool:
        """Write a message to the Vestaboard.

        `message` must be a two-dimensional (6, 22) array of character codes
        representing the exact positions of characters on the board.

        :raises ValueError: if ``message`` is a list with unsupported dimensions
        """
        validate_rows(message)
        r = self.session.post("", json=message)
        r.raise_for_status()
        return r.status_code == requests.codes.CREATED
